#include <funkeymonkey/funkeymonkeymodule.h>
#include <funkeymonkey/uinputdevice.h>

#include <iostream>
#include <thread>
#include <chrono>
#include <mutex>
#include <condition_variable>
#include <regex>
#include <fstream>
#include <algorithm>
#include <cctype>
#include <unordered_map>
#include <functional>
#include <array>
#include <stdlib.h> 

enum Role { ROLE_LEFT_NUB, ROLE_RIGHT_NUB, ROLE_KEYBOARD, ROLE_GPIO };

struct LeftRight {
	bool left	= false;
	bool right	= false;
	bool pressed	= false;
};

struct Scripts {
	std::string brightnessNormal;
	std::string brightnessFn;
	std::string brightnessShift;
	std::string brightnessFnShift;
	std::string brightnessAlt;
	std::string brightnessCtrl;
	std::string brightnessAltCtrl;
	std::string brightnessFnAlt;
	std::string brightnessFnCtrl;
	std::string brightnessShiftAlt;
	std::string brightnessShiftCtrl;
	std::string brightnessFnShiftAlt;
	std::string brightnessFnShiftCtrl;
        std::string commandModeOn;
        std::string commandModeOff;
};

static constexpr unsigned int FIRST_KEY = KEY_RESERVED;
static constexpr unsigned int LAST_KEY = KEY_UNKNOWN;

template<int FIRST_KEY, int LAST_KEY> class KeyBehaviors {
public:
	void passthrough(unsigned int code);
	void map(unsigned int code, unsigned int result);
	void gphat(unsigned int code, unsigned int gamepad);
	void gpmap(unsigned int code, unsigned int gamepad);
	void gpmap2(unsigned int code, unsigned int gamepad, bool* self, LeftRight* pair);
	void altmap(unsigned int code, bool* flag, unsigned int regular, unsigned int alternative);
        void altcomplex(unsigned int code, bool* flag, unsigned int regular, std::function<void(int)> function);
	void complex(unsigned int code, std::function<void(int)> function);
	void handle(unsigned int code, int value);

private:
	struct KeyBehavior {
		KeyBehavior() : type(PASSTHROUGH), mapping(0), function() {}
		enum Type  { PASSTHROUGH, MAPPED, ALTMAPPED, COMPLEX, ALTCOMPLEX, GPMAPPED,GPMAP2, GPHAT };
		Type type;
		int mapping;
		int alternative;
		bool pressed_as;
		bool* flag;
		LeftRight* lr;
		std::function<void(int)> function;
	};

	static constexpr unsigned int NUM_KEYS = LAST_KEY - FIRST_KEY + 1;
	std::array<KeyBehavior, NUM_KEYS> behaviors;
};

struct Mouse {
	// Any thread using device must hold mutex
	UinputDevice device;
	
	// May be changed from any thread at any time
	int dx;
	int dy;
	int dwx;
	int dwy;
	
        // Mouse wheel position
        int wheelPosition;
        int hwheelPosition;
        static int const WHEEL_HI_RES_PER_NOTCH = 120;

	// Used to signal changes in above values, uses mutex
	std::condition_variable signal;
	
	// Mouse device mutex
	std::mutex mutex;
};


struct Settings {
	enum NubAxisMode {
		UNKNOWN_NUB_AXIS_MODE,
		MOUSE_X, MOUSE_Y, MOUSE_BTN,
		SCROLL_X, SCROLL_Y
	};
	enum NubClickMode {
		UNKNOWN_NUB_CLICK_MODE,
		NUB_CLICK_LEFT, NUB_CLICK_RIGHT,
		MOUSE_LEFT, MOUSE_MIDDLE, MOUSE_RIGHT
	};
	enum AltMode {
		LEFT_ALT, RIGHT_ALT
	};
        enum ProfileIndex {
                PROFILE_0, PROFILE_1, PROFILE_2, PROFILE_3, PROFILE_4,
                PROFILE_5, PROFILE_6, PROFILE_7, PROFILE_8, PROFILE_9,
                PROFILE_NONE
        };
	
        std::vector<NubAxisMode> leftNubModeX = { MOUSE_X };
        std::vector<NubAxisMode> leftNubModeY = { MOUSE_Y };
        std::vector<NubAxisMode> rightNubModeX = { MOUSE_BTN };
        std::vector<NubAxisMode> rightNubModeY = { SCROLL_Y };
        std::vector<NubClickMode> leftNubClickMode = { MOUSE_LEFT };
        std::vector<NubClickMode> rightNubClickMode = { MOUSE_RIGHT };
    
        std::vector<AltMode> startButton  = { LEFT_ALT };
        std::vector<AltMode> rightTrigger = { RIGHT_ALT };
	
	// May be changed from any thread at any time
	int joyDeadzone = 10;
	int mouseDeadzone = 50;
	int mouseSensitivity = 75;
        int mouseWheelSensitivity = 75;
	int mouseWheelDeadzone = 100;
	int mouseClickDeadzone = 100;
	
	bool exportGamepad = true;
	bool exportMouse   = true;
	bool exportKeypad  = true;

	std::string configFile;
        std::string profiles[10];
        ProfileIndex currentProfile = PROFILE_NONE;
	Scripts scripts;
};

using AltModeMap = std::unordered_map<std::string, Settings::AltMode>;
AltModeMap const ALT_MODES = {
	{ "left_alt",  Settings::LEFT_ALT },
	{ "right_alt", Settings::RIGHT_ALT }
};

using NubAxisModeMap = std::unordered_map<std::string, Settings::NubAxisMode>;
NubAxisModeMap const NUB_AXIS_MODES = {
	{ "mouse_x", Settings::MOUSE_X },
	{ "mouse_y", Settings::MOUSE_Y },
	{ "mouse_btn", Settings::MOUSE_BTN },
	{ "scroll_x", Settings::SCROLL_X },
	{ "scroll_y", Settings::SCROLL_Y }
};

using NubClickModeMap = std::unordered_map<std::string, Settings::NubClickMode>;
NubClickModeMap const NUB_CLICK_MODES = {
	{ "nub_click_left", Settings::NUB_CLICK_LEFT },
	{ "nub_click_right", Settings::NUB_CLICK_RIGHT },
	{ "mouse_left", Settings::MOUSE_LEFT },
	{ "mouse_middle", Settings::MOUSE_MIDDLE },
	{ "mouse_right", Settings::MOUSE_RIGHT }
};

void handleArgs(char const** argv, unsigned int argc, Settings& settings);
void loadConfig(std::string const& filename, Settings& settings);
void loadProfile(Settings::ProfileIndex index, Settings& settings);
void setCommandMode(bool value);
Settings::NubAxisMode parseNubAxisMode(std::string const& str);
Settings::NubClickMode parseNubClickMode(std::string const& str);
Settings::AltMode parseAltMode(std::string const& str);
void handleNubAxis(Settings::NubAxisMode mode, int value, Mouse* mouse, UinputDevice* gamepad, Settings const& settings);
void handleNubClick(Settings::NubClickMode mode, int value, Mouse* mouse, UinputDevice* gamepad, Settings const& settings);
void handleBrightness(int value);

// Mouse movement/scroll thread handler
void handleMouse(Mouse* mouse, Settings* settings, bool* stop);

// Parse whitespace separated list of items
template<class T> std::vector<T> parseMultivalue(std::string const& str, std::function<T(std::string const&)> fn);

struct {
	bool stop = false;
	UinputDevice* gamepad = nullptr;
	UinputDevice* keyboard = nullptr;
	KeyBehaviors<FIRST_KEY, LAST_KEY>* behaviors = nullptr;
	Mouse* mouse = nullptr;
	std::thread mouseThread;
	Settings settings;
	LeftRight Fn;
	LeftRight Alt;
	LeftRight Shift;
	LeftRight Ctrl;
	int hatx = 0;
	int haty = 0;
	int mouseBtn = 0;
        bool commandMode = false;
} global;


void init(char const** argv, unsigned int argc) {
	std::vector<unsigned int> keycodes;
	for(unsigned int i = FIRST_KEY; i <= LAST_KEY; ++i) {
		keycodes.push_back(i);
	}
	
	global.keyboard = new UinputDevice("/dev/uinput", BUS_USB, "pyraInput keyboard", 1, 1, 1, {
		{ EV_KEY, keycodes }
	});

	global.behaviors = new KeyBehaviors<FIRST_KEY, LAST_KEY>();

        global.behaviors->altcomplex(KEY_MENU, &global.Fn.pressed, KEY_MENU, [](int value) {
            if(value) {
              setCommandMode(!global.commandMode);
            }
        });

	global.behaviors->gphat(KEY_UP,		BTN_DPAD_UP);
	global.behaviors->gphat(KEY_DOWN,	BTN_DPAD_DOWN);
	global.behaviors->gphat(KEY_LEFT,	BTN_DPAD_LEFT);
	global.behaviors->gphat(KEY_RIGHT,	BTN_DPAD_RIGHT);
	global.behaviors->gpmap2(KEY_LEFTALT,	BTN_START, &global.Alt.left, &global.Alt);
	global.behaviors->gpmap2(KEY_LEFTCTRL,	BTN_SELECT, &global.Ctrl.left, &global.Ctrl);
	global.behaviors->gpmap(KEY_HOME,	BTN_A);
	global.behaviors->gpmap(KEY_END,	BTN_B);
	global.behaviors->gpmap(KEY_PAGEDOWN,	BTN_X);
	global.behaviors->gpmap(KEY_PAGEUP,	BTN_Y);
	global.behaviors->gpmap2(KEY_RIGHTSHIFT,BTN_TL, &global.Shift.right, &global.Shift);
	global.behaviors->gpmap2(KEY_RIGHTCTRL,	BTN_TR, &global.Ctrl.right, &global.Ctrl);
	global.behaviors->gpmap2(KEY_RIGHTALT,	BTN_TR2, &global.Alt.right, &global.Alt);
	global.behaviors->gpmap(KEY_INSERT,	BTN_C);  //(I)
	global.behaviors->gpmap(KEY_DELETE,	BTN_Z);  //(II)
	
	global.behaviors->complex(KEY_LEFTSHIFT, [](int value) {
		global.Shift.left = (value==1);
		global.Shift.pressed = global.Shift.left || global.Shift.right;
		global.keyboard->send(EV_KEY, KEY_LEFTSHIFT, value);
	});
	global.behaviors->complex(KEY_RIGHTMETA, [](int value) {
		global.Fn.right = (value==1);
		global.Fn.pressed = global.Fn.left || global.Fn.right;
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_KEY, BTN_TL2, value);
			global.gamepad->send(EV_SYN, 0, 0);
		}
	});
	global.behaviors->complex(KEY_LEFTMETA, [](int value) {
		global.Fn.left = (value==1);
		global.Fn.pressed = global.Fn.left || global.Fn.right;
	});

	global.behaviors->altmap(KEY_ESC,	&global.Fn.pressed, KEY_ESC,	KEY_SYSRQ);
	global.behaviors->altmap(KEY_PAUSE,	&global.Fn.pressed, KEY_PAUSE,	KEY_SCALE);
        global.behaviors->complex(KEY_BRIGHTNESSUP, handleBrightness);
	//global.behaviors->altmap(KEY_BRIGHTNESSUP,	&global.Fn.pressed, KEY_BRIGHTNESSUP,	KEY_BRIGHTNESSDOWN);
	global.behaviors->altmap(KEY_F11,	&global.Fn.pressed, KEY_F11,	KEY_F12);

        // Handler function generator function for number keys
        auto numKeyBehavior = [](unsigned int regular, unsigned int fn, Settings::ProfileIndex index) {
          return [regular, fn, index](int value) {
            if(global.commandMode) {
              if(value) {
                loadProfile(index, global.settings);
                setCommandMode(false);
              }
            } else if(global.Fn.pressed) {
              global.keyboard->send(EV_KEY, fn, value);
            } else {
              global.keyboard->send(EV_KEY, regular, value);
            }
          };
        };

	global.behaviors->complex(KEY_1, numKeyBehavior(KEY_1, KEY_F1, Settings::PROFILE_1));
	global.behaviors->complex(KEY_2, numKeyBehavior(KEY_2, KEY_F2, Settings::PROFILE_2));
	global.behaviors->complex(KEY_3, numKeyBehavior(KEY_3, KEY_F3, Settings::PROFILE_3));
	global.behaviors->complex(KEY_4, numKeyBehavior(KEY_4, KEY_F4, Settings::PROFILE_4));
	global.behaviors->complex(KEY_5, numKeyBehavior(KEY_5, KEY_F5, Settings::PROFILE_5));
	global.behaviors->complex(KEY_6, numKeyBehavior(KEY_6, KEY_F6, Settings::PROFILE_6));
	global.behaviors->complex(KEY_7, numKeyBehavior(KEY_7, KEY_F7, Settings::PROFILE_7));
	global.behaviors->complex(KEY_8, numKeyBehavior(KEY_8, KEY_F8, Settings::PROFILE_8));
	global.behaviors->complex(KEY_9, numKeyBehavior(KEY_9, KEY_F9, Settings::PROFILE_9));
	global.behaviors->complex(KEY_0, numKeyBehavior(KEY_0, KEY_F10, Settings::PROFILE_0));

	global.behaviors->altmap(KEY_TAB,	&global.Fn.pressed, KEY_TAB,	KEY_CAPSLOCK);
	global.behaviors->altmap(KEY_Q,		&global.Fn.pressed, KEY_Q,	KEY_MACRO);			// I120 // ok
	global.behaviors->altmap(KEY_W,		&global.Fn.pressed, KEY_W,	KEY_KPCOMMA);			// I129 // ok
	global.behaviors->altmap(KEY_E,		&global.Fn.pressed, KEY_E,	KEY_SETUP);			// I149 // ok
	global.behaviors->altmap(KEY_R,		&global.Fn.pressed, KEY_R,	KEY_DELETEFILE);		// I154 // ok
	global.behaviors->altmap(KEY_T,		&global.Fn.pressed, KEY_T,	KEY_CLOSECD);			// I168 // ok
	global.behaviors->altmap(KEY_Y,		&global.Fn.pressed, KEY_Y,	KEY_ISO);			// I178 // ok
	global.behaviors->altmap(KEY_U,		&global.Fn.pressed, KEY_U,	KEY_MOVE);			// I183 // ok
	global.behaviors->altmap(KEY_I,		&global.Fn.pressed, KEY_I,	KEY_EDIT);			// I184 // ok
	global.behaviors->altmap(KEY_O,		&global.Fn.pressed, KEY_O,	KEY_ALTERASE);			// I230 // ok
	global.behaviors->altmap(KEY_P,		&global.Fn.pressed, KEY_P,	KEY_BASSBOOST);			// I217 // ok
	global.behaviors->altmap(KEY_APOSTROPHE,	&global.Fn.pressed, KEY_APOSTROPHE,	KEY_UWB);	// I247 // ok
	global.behaviors->altmap(KEY_A,		&global.Fn.pressed, KEY_A,	KEY_QUESTION);			// I222 // ok
	global.behaviors->altmap(KEY_S,		&global.Fn.pressed, KEY_S,	KEY_UNKNOWN);			// I248 // ok
	global.behaviors->altmap(KEY_D,		&global.Fn.pressed, KEY_D,	KEY_SOUND);			// I221 // ok
	global.behaviors->altmap(KEY_F,		&global.Fn.pressed, KEY_F,	KEY_HP);			// I219 // ok
	global.behaviors->altmap(KEY_G,		&global.Fn.pressed, KEY_G,	KEY_RO);			// I249- AB11 (89)
	global.behaviors->altmap(KEY_H,		&global.Fn.pressed, KEY_H,	KEY_KPJPCOMMA);			// I250- JPCM (95)
	global.behaviors->altmap(KEY_J,		&global.Fn.pressed, KEY_J,	KEY_YEN);			// I251- AE13 (124)
	global.behaviors->altmap(KEY_K,		&global.Fn.pressed, KEY_K,	KEY_F19);			// I252- FK19 (189)
	global.behaviors->altmap(KEY_L,		&global.Fn.pressed, KEY_L,	KEY_F24);			// I253- FK24 (194)
	global.behaviors->altmap(KEY_COMMA,	&global.Fn.pressed, KEY_COMMA,	KEY_SEMICOLON);
	global.behaviors->altmap(KEY_DOT,	&global.Fn.pressed, KEY_DOT,	KEY_SLASH);
	global.behaviors->altmap(KEY_Z,		&global.Fn.pressed, KEY_Z,	KEY_EQUAL);
	global.behaviors->altmap(KEY_X,		&global.Fn.pressed, KEY_X,	KEY_MINUS);
	global.behaviors->altmap(KEY_C,		&global.Fn.pressed, KEY_C,	KEY_LEFTBRACE);
	global.behaviors->altmap(KEY_V,		&global.Fn.pressed, KEY_V,	KEY_RIGHTBRACE);
	global.behaviors->altmap(KEY_B,		&global.Fn.pressed, KEY_B,	KEY_BACKSLASH);
	global.behaviors->altmap(KEY_N,		&global.Fn.pressed, KEY_N,	KEY_GRAVE);
	global.behaviors->altmap(KEY_M,		&global.Fn.pressed, KEY_M,	195);			// 228- MDSW (195)
	global.behaviors->altmap(KEY_SPACE,	&global.Fn.pressed, KEY_SPACE,	KEY_COMPOSE);
	
	global.gamepad = new UinputDevice("/dev/uinput", BUS_USB, "pyraInput Gamepad", 1, 1, 1, {
		{ EV_KEY, {
			BTN_A, BTN_B, BTN_X, BTN_Y, 
			BTN_TL, BTN_TR, BTN_TL2, BTN_TR2,
			BTN_SELECT, BTN_START, BTN_C, BTN_Z,
			BTN_THUMBL, BTN_THUMBR
		} },
		{ EV_ABS, { ABS_HAT0X, ABS_HAT0Y, ABS_X, ABS_Y, ABS_RX, ABS_RY } }
	}, {
                { ABS_X, -255, 255, 3, 30 },
                { ABS_Y, -255, 255, 3, 30 },
                { ABS_RX, -255, 255, 3, 30 },
                { ABS_RY, -255, 255, 3, 30 },
                { ABS_HAT0X, -65535, 65535, 0, 0 },
                { ABS_HAT0Y, -65535, 65535, 0, 0 },
        });
	global.mouse = new Mouse {
		UinputDevice("/dev/uinput", BUS_USB, "pyraInput Mouse", 1, 1, 1, {
			{ EV_KEY, { BTN_LEFT, BTN_RIGHT } },
	       { EV_REL, { REL_X, REL_Y, REL_HWHEEL, REL_WHEEL, REL_WHEEL_HI_RES, REL_HWHEEL_HI_RES } }
		}), 0, 0, 0, 0, {}, {}
	};
	global.mouseThread = std::move(std::thread(handleMouse, global.mouse, &global.settings, &global.stop));

	handleArgs(argv, argc, global.settings);
	
	global.settings.scripts.brightnessNormal = "/usr/share/pyra/scripts/pyra-brightkey.sh screen up";
	global.settings.scripts.brightnessShift = "/usr/share/pyra/scripts/pyra-brightkey.sh screen down";
	global.settings.scripts.brightnessFn = "/usr/share/pyra/scripts/pyra-brightkey.sh key up";
	global.settings.scripts.brightnessFnShift = "/usr/share/pyra/scripts/pyra-brightkey.sh key down";

	if(!global.settings.configFile.empty()) {
		loadConfig(global.settings.configFile, global.settings);
	}
}

void handle(input_event const& e, unsigned int role) {
	switch(e.type) {
	case EV_ABS:
		if (role == ROLE_LEFT_NUB) {
			switch(e.code) {
			case ABS_X:
				if (global.settings.exportGamepad && (e.value>global.settings.joyDeadzone||e.value<-global.settings.joyDeadzone)) {
					global.gamepad->send(EV_ABS, ABS_X, e.value);
					global.gamepad->send(EV_SYN, 0, 0);
				}
                                for(auto mode : global.settings.leftNubModeX) {
                                  handleNubAxis(mode, e.value, global.mouse, global.gamepad, global.settings);
                                }
				break;
			case ABS_Y:
				if (global.settings.exportGamepad && (e.value>global.settings.joyDeadzone||e.value<-global.settings.joyDeadzone)) {
					global.gamepad->send(EV_ABS, ABS_Y, e.value);
					global.gamepad->send(EV_SYN, 0, 0);
				}
                                for(auto mode : global.settings.leftNubModeY) {
                                  handleNubAxis(mode, e.value, global.mouse, global.gamepad, global.settings);
                                }
				break;
			default: break;
			};
		} else if (role == ROLE_RIGHT_NUB) {
			switch(e.code) {
			case ABS_X:
				if (global.settings.exportGamepad && (e.value>global.settings.joyDeadzone||e.value<-global.settings.joyDeadzone)) {
					global.gamepad->send(EV_ABS, ABS_RX, e.value);
					global.gamepad->send(EV_SYN, 0, 0);
				}
                                for(auto mode : global.settings.rightNubModeX) {
                                  handleNubAxis(mode, e.value, global.mouse, global.gamepad, global.settings);
                                }
				break;
			case ABS_Y:
				if (global.settings.exportGamepad && (e.value>global.settings.joyDeadzone||e.value<-global.settings.joyDeadzone)) {
					global.gamepad->send(EV_ABS, ABS_RY, e.value);
					global.gamepad->send(EV_SYN, 0, 0);
				}
                                for(auto mode : global.settings.rightNubModeY) {
                                  handleNubAxis(mode, e.value, global.mouse, global.gamepad, global.settings);
                                }
				break;
			default: break;
		};
		}
		break;
	case EV_KEY:
		switch(e.code) {
                case BTN_LEFT: // Right nub click
                    if (role == ROLE_GPIO) {
                        for(auto mode : global.settings.rightNubClickMode) {
                          handleNubClick(mode, e.value, global.mouse, global.gamepad, global.settings);
                        }
                    }
                    break;
		case BTN_RIGHT: // Left nub click
                    if (role == ROLE_GPIO) {
                        for(auto mode : global.settings.leftNubClickMode) {
                          handleNubClick(mode, e.value, global.mouse, global.gamepad, global.settings);
                        }
                    }
                    break;

		default: 
			global.behaviors->handle(e.code, e.value);
			global.keyboard->send(EV_SYN, 0, 0);
			break;
		}
		break;
	case EV_REL:
		/* 
		 * One of the nub is in mouse mode
		 * TODO: Should forward the event to the mouse
		global.mouse->device.send(EV_REL, e.code, e.value);
		global.mouse->signal.notify_all();
		printf("REL: type=%d, code=%d, value=%d\n", e.type, e.code, e.value);
		 */
		break;
	}
}

void destroy() {
	global.stop = true;
	global.mouse->signal.notify_all();

	if(global.gamepad) {
		delete global.gamepad;
	}
	if(global.keyboard) {
		delete global.keyboard;
	}
	if(global.behaviors) {
		delete global.behaviors;
	}

	global.mouseThread.join();
}

void user1() {
	loadProfile(Settings::PROFILE_NONE, global.settings);
	loadConfig(global.settings.configFile, global.settings);
}
void user2() {
}

void handleArgs(char const** argv, unsigned int argc, Settings& settings) {
	std::string configFile;
	std::regex configRegex("config=(.*)");
	
	for(unsigned int i = 0; i < argc; ++i) {
		std::string arg(argv[i]);
		std::smatch match;
		if(std::regex_match(arg, match, configRegex)) {
			settings.configFile = match[1];
		}
	}
}

using SettingHandler = std::function<void(std::string const&,Settings&)>;
using SettingHandlerMap = std::unordered_map<std::string, SettingHandler>;
SettingHandlerMap const SETTING_HANDLERS = {
	{ "scripts.brightness.normal", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessNormal = value;
	} },
	{ "scripts.brightness.fn", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFn = value;
	} },
	{ "scripts.brightness.shift", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShift = value;
	} },
	{ "scripts.brightness.fnshift", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShift = value;
	} },
	{ "scripts.brightness.alt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessAlt = value;
	} },
	{ "scripts.brightness.ctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessCtrl = value;
	} },
	{ "scripts.brightness.altctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessAltCtrl = value;
	} },
	{ "scripts.brightness.fnalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnAlt = value;
	} },
	{ "scripts.brightness.fnctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnCtrl = value;
	} },
	{ "scripts.brightness.shiftalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShiftAlt = value;
	} },
	{ "scripts.brightness.shiftctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessShiftCtrl = value;
	} },
	{ "scripts.brightness.fnshiftalt", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShiftAlt = value;
	} },
	{ "scripts.brightness.fnshiftctrl", [](std::string const& value, Settings& settings){
		settings.scripts.brightnessFnShiftCtrl = value;
	} },
	{ "scripts.commandmode.on", [](std::string const& value, Settings& settings){
		settings.scripts.commandModeOn = value;
	} },
	{ "scripts.commandmode.off", [](std::string const& value, Settings& settings){
		settings.scripts.commandModeOff = value;
	} },
	{ "profile.0", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_0] = value;
	} },
	{ "profile.1", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_1] = value;
	} },
	{ "profile.2", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_2] = value;
	} },
	{ "profile.3", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_3] = value;
	} },
	{ "profile.4", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_4] = value;
	} },
	{ "profile.5", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_5] = value;
	} },
	{ "profile.6", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_6] = value;
	} },
	{ "profile.7", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_7] = value;
	} },
	{ "profile.8", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_8] = value;
	} },
	{ "profile.9", [](std::string const& value, Settings& settings){
		settings.profiles[Settings::PROFILE_9] = value;
	} },
	{ "profile.current", [](std::string const& value, Settings& settings){
		int intValue = std::stoi(value);

                if(intValue >= 0 && intValue < 10) {
                  loadProfile(static_cast<Settings::ProfileIndex>(intValue), settings);
                }
	} },
	{ "gamepad.export", [](std::string const& value, Settings& settings){
		settings.exportGamepad = (value != "0");
	} },
	{ "keypad.export", [](std::string const& value, Settings& settings){
		settings.exportKeypad = (value != "0");
	} },
	{ "mouse.export", [](std::string const& value, Settings& settings){
		settings.exportMouse = (value != "0");
	} },
	{ "mouse.sensitivity", [](std::string const& value, Settings& settings){
		settings.mouseSensitivity = std::stoi(value);
	} },
	{ "mouse.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseDeadzone = std::stoi(value);
	} },
	{ "mouse.wheel.sensitivity", [](std::string const& value, Settings& settings){
		settings.mouseWheelSensitivity = std::stoi(value);
	} },
	{ "mouse.wheel.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseWheelDeadzone = std::stoi(value);
	} },
	{ "mouse.click.deadzone", [](std::string const& value, Settings& settings) {
		settings.mouseClickDeadzone = std::stoi(value);
	} },
	{ "nubs.deadzone", [](std::string const& value, Settings& settings) {
		settings.joyDeadzone = std::stoi(value);
	} },
	{ "nubs.left.x", [](std::string const& value, Settings& settings) {
                settings.leftNubModeX = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.left.y", [](std::string const& value, Settings& settings) {
		settings.leftNubModeY = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.right.x", [](std::string const& value, Settings& settings) {
		settings.rightNubModeX = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.right.y", [](std::string const& value, Settings& settings) {
		settings.rightNubModeY = parseMultivalue<Settings::NubAxisMode>(value, parseNubAxisMode);
	} },
	{ "nubs.left.click", [](std::string const& value, Settings& settings) {
		settings.leftNubClickMode = parseMultivalue<Settings::NubClickMode>(value, parseNubClickMode);
	} },
	{ "nubs.right.click", [](std::string const& value, Settings& settings) {
		settings.rightNubClickMode = parseMultivalue<Settings::NubClickMode>(value, parseNubClickMode);
	} },
	{ "altmode.start.button", [](std::string const& value, Settings& settings) {
		settings.startButton = parseMultivalue<Settings::AltMode>(value, parseAltMode);
	} },
	{ "altmode.right.trigger", [](std::string const& value, Settings& settings) {
		settings.rightTrigger = parseMultivalue<Settings::AltMode>(value, parseAltMode);
	} }
};

void loadConfig(std::string const& filename, Settings& settings) {
	std::regex re("^([\\w.]+)\\s*=\\s*(.*)$");
	std::regex emptyRe("^\\s*$");
	std::ifstream configFile(filename);
	
	if(!configFile) {
		std::cerr << "ERROR: Could not open config file " << filename << std::endl;
		return;
	}
	std::string line;
	while(std::getline(configFile, line)) {
		if(line.empty() || line.at(0) == '#' || std::regex_match(line, emptyRe))
			continue;

		std::smatch match;
		if(std::regex_match(line, match, re)) {
			std::string key = match[1];
			std::string value = match[2];
			std::transform(key.begin(), key.end(), key.begin(), tolower);
			auto iter = SETTING_HANDLERS.find(key);
			if(iter == SETTING_HANDLERS.end()) {
				std::cout << "WARNING: Unknown setting in config file: " 
				<< key << std::endl;
			} else {
				iter->second(value, settings);
			}
		} else {
			std::cerr << "Invalid line in config file: " << line << std::endl;
		}
	}
	/* Apply "Alt Keys" mapping configuration */
    if (std::find(settings.startButton.begin(), settings.startButton.end(), Settings::LEFT_ALT) != settings.startButton.end()) {
        global.behaviors->gpmap2(KEY_LEFTALT,	BTN_START, &global.Alt.left, &global.Alt);
    }
    if (std::find(settings.startButton.begin(), settings.startButton.end(), Settings::RIGHT_ALT) != settings.startButton.end()) {
        global.behaviors->gpmap2(KEY_RIGHTALT,	BTN_START, &global.Alt.right, &global.Alt);
    }
    if (std::find(settings.rightTrigger.begin(), settings.rightTrigger.end(), Settings::LEFT_ALT) != settings.rightTrigger.end()) {
        global.behaviors->gpmap2(KEY_LEFTALT,	BTN_TR2, &global.Alt.left, &global.Alt);
    }
    if (std::find(settings.rightTrigger.begin(), settings.rightTrigger.end(), Settings::RIGHT_ALT) != settings.rightTrigger.end()) {
        global.behaviors->gpmap2(KEY_RIGHTALT,	BTN_TR2, &global.Alt.right, &global.Alt);
    }
}

void loadProfile(Settings::ProfileIndex index, Settings& settings) {
  if(index != settings.currentProfile) {
    if(index != Settings::PROFILE_NONE) {
      std::string const& profilePath = settings.profiles[index];
      loadConfig(profilePath, settings);
    }
    settings.currentProfile = index;
  }
}

void setCommandMode(bool value) {
  if(global.commandMode != value) {
    global.commandMode = value;

    if(global.commandMode) {
      if(!global.settings.scripts.commandModeOn.empty()) {
        system(global.settings.scripts.commandModeOn.c_str());
      }
    } else {
      if(!global.settings.scripts.commandModeOff.empty()) {
        system(global.settings.scripts.commandModeOff.c_str());
      }
    }
  }
}

Settings::NubAxisMode parseNubAxisMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = NUB_AXIS_MODES.find(s);
	if(iter == NUB_AXIS_MODES.end()) {
		return Settings::UNKNOWN_NUB_AXIS_MODE;
	} else {
		return iter->second;
	}
}

Settings::NubClickMode parseNubClickMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = NUB_CLICK_MODES.find(s);
	if(iter == NUB_CLICK_MODES.end()) {
		return Settings::UNKNOWN_NUB_CLICK_MODE;
	} else {
		return iter->second;
	}
}

Settings::AltMode parseAltMode(std::string const& str) {
	std::string s(str);
	std::transform(s.begin(), s.end(), s.begin(), tolower);
	auto iter = ALT_MODES.find(s);
	if(iter == ALT_MODES.end()) {
		return Settings::LEFT_ALT;
	} else {
		return iter->second;
	}
}

void handleNubAxis(Settings::NubAxisMode mode, int value, Mouse* mouse, UinputDevice* gamepad, Settings const& settings) {
	switch(mode) {
	case Settings::MOUSE_X:
		mouse->dx = value;
		if(mouse->dx > settings.mouseDeadzone || mouse->dx < -settings.mouseDeadzone)
			mouse->signal.notify_all();
		break;
	case Settings::MOUSE_Y:
		mouse->dy = value;
		if(mouse->dy > settings.mouseDeadzone || mouse->dy < -settings.mouseDeadzone)
			mouse->signal.notify_all();
		break;
	case Settings::SCROLL_X:
		mouse->dwx = value;
		if(mouse->dwx > settings.mouseDeadzone || mouse->dwx < -settings.mouseDeadzone)
			mouse->signal.notify_all();
		break;
	case Settings::SCROLL_Y:
		mouse->dwy = value;
		if(mouse->dwy > settings.mouseDeadzone  || mouse->dwy < -settings.mouseDeadzone)
			mouse->signal.notify_all();
		break;
	case Settings::MOUSE_BTN: {
		int new_val = 0;
		if (value < -settings.mouseClickDeadzone) 
			new_val = -1;
		else if (value > settings.mouseClickDeadzone) 
			new_val = 1;
		if (global.mouseBtn!=new_val && global.settings.exportMouse) {
			if (global.mouseBtn == -1) 
				mouse->device.send(EV_KEY, BTN_LEFT, 0);
			else if (global.mouseBtn == 1) 
				mouse->device.send(EV_KEY, BTN_RIGHT, 0);
			if (new_val == -1 )
				mouse->device.send(EV_KEY, BTN_LEFT, 1);
			else if (new_val == 1 )
				mouse->device.send(EV_KEY, BTN_RIGHT, 1);
			mouse->device.send(EV_SYN, 0, 0);
			global.mouseBtn=new_val;
		}
		}
		break;
	case Settings::UNKNOWN_NUB_AXIS_MODE:
		break;
	}
}

void handleNubClick(Settings::NubClickMode mode, int value, Mouse* mouse, UinputDevice* gamepad, Settings const& settings) {
	switch(mode) {
	case Settings::MOUSE_MIDDLE: {
		if (global.settings.exportMouse) {
			std::lock_guard<std::mutex> lk(mouse->mutex);
			mouse->device.send(EV_KEY, BTN_MIDDLE, value);
			mouse->device.send(EV_SYN, 0, 0);
		}
		break;
	}
	case Settings::MOUSE_LEFT: {
		if (global.settings.exportMouse) {
			std::lock_guard<std::mutex> lk(mouse->mutex);
			mouse->device.send(EV_KEY, BTN_LEFT, value);
			mouse->device.send(EV_SYN, 0, 0);
		}
		break;
	}
	case Settings::MOUSE_RIGHT: {
		if (global.settings.exportMouse) {
			std::lock_guard<std::mutex> lk(mouse->mutex);
			mouse->device.send(EV_KEY, BTN_RIGHT, value);
			mouse->device.send(EV_SYN, 0, 0);
		}
		break;
	}
	case Settings::NUB_CLICK_LEFT:
		if (global.settings.exportGamepad) {
			gamepad->send(EV_KEY, BTN_THUMBL, value);
			gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case Settings::NUB_CLICK_RIGHT:
		if (global.settings.exportGamepad) {
			gamepad->send(EV_KEY, BTN_THUMBR, value);
			gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case Settings::UNKNOWN_NUB_CLICK_MODE:
		break;
	}
}

void handleBrightness(int value) {
  if (value!=1) return;
  int ret = 0;
  if (global.Shift.pressed && global.Ctrl.pressed && global.Fn.pressed) {
    if (!global.settings.scripts.brightnessFnShiftCtrl.empty())
      ret=system(global.settings.scripts.brightnessFnShiftCtrl.c_str());
  } else if (global.Shift.pressed && global.Alt.pressed && global.Fn.pressed) {
    if (!global.settings.scripts.brightnessFnShiftAlt.empty())
      ret=system(global.settings.scripts.brightnessFnShiftAlt.c_str());
  } else if (global.Shift.pressed && global.Alt.pressed) {
    if (!global.settings.scripts.brightnessShiftAlt.empty())
      ret=system(global.settings.scripts.brightnessShiftAlt.c_str());
  } else if (global.Shift.pressed && global.Ctrl.pressed) {
    if (!global.settings.scripts.brightnessShiftCtrl.empty())
      ret=system(global.settings.scripts.brightnessShiftCtrl.c_str());
  } else if (global.Fn.pressed && global.Alt.pressed) {
    if (!global.settings.scripts.brightnessFnAlt.empty())
      ret=system(global.settings.scripts.brightnessFnAlt.c_str());
  } else if (global.Fn.pressed && global.Ctrl.pressed) {
    if (!global.settings.scripts.brightnessFnCtrl.empty())
      ret=system(global.settings.scripts.brightnessFnCtrl.c_str());
  } else if (global.Alt.pressed && global.Ctrl.pressed) {
    if (!global.settings.scripts.brightnessAltCtrl.empty())
      ret=system(global.settings.scripts.brightnessAltCtrl.c_str());
  } else if (global.Fn.pressed && global.Shift.pressed) {
    if (!global.settings.scripts.brightnessFnShift.empty())
      ret=system(global.settings.scripts.brightnessFnShift.c_str());
  } else if (global.Alt.pressed) {
    if (!global.settings.scripts.brightnessAlt.empty())
      ret=system(global.settings.scripts.brightnessAlt.c_str());
  } else if (global.Alt.pressed) {
    if (!global.settings.scripts.brightnessAlt.empty())
      ret=system(global.settings.scripts.brightnessAlt.c_str());
  } else if (global.Shift.pressed) {
    if (!global.settings.scripts.brightnessShift.empty())
      ret=system(global.settings.scripts.brightnessShift.c_str());
  } else if (global.Fn.pressed) {
    if (!global.settings.scripts.brightnessFn.empty())
      ret=system(global.settings.scripts.brightnessFn.c_str());
  } else {
    if (!global.settings.scripts.brightnessNormal.empty())
      ret=system(global.settings.scripts.brightnessNormal.c_str());
  }
}

void handleMouse(Mouse* mouse, Settings* settings, bool* stop) {
	while(!*stop) {
		if((mouse->dx > settings->mouseDeadzone || mouse->dx < -settings->mouseDeadzone || mouse->dy > settings->mouseDeadzone || mouse->dy < -settings->mouseDeadzone || mouse->dwx > settings->mouseClickDeadzone || mouse->dwx < -settings->mouseClickDeadzone || mouse->dwy > settings->mouseWheelDeadzone || mouse->dwy < -settings->mouseWheelDeadzone)&&global.settings.exportMouse) {
			std::lock_guard<std::mutex> lk(mouse->mutex);

			if(mouse->dx > settings->mouseDeadzone) {
				mouse->device.send(EV_REL, REL_X, (mouse->dx - settings->mouseDeadzone) * settings->mouseSensitivity / 1000);
			} else if(mouse->dx < -settings->mouseDeadzone) {
				mouse->device.send(EV_REL, REL_X, (mouse->dx + settings->mouseDeadzone) * settings->mouseSensitivity / 1000);
			}

			if(mouse->dy > settings->mouseDeadzone) {
				mouse->device.send(EV_REL, REL_Y, (mouse->dy - settings->mouseDeadzone) * settings->mouseSensitivity / 1000);
			} else if(mouse->dy < -settings->mouseDeadzone) {
				mouse->device.send(EV_REL, REL_Y, (mouse->dy + settings->mouseDeadzone) * settings->mouseSensitivity / 1000);
			}


                        int hwheelChange = 0;
			if(mouse->dwx > settings->mouseWheelDeadzone) {
                            hwheelChange = (mouse->dwx - settings->mouseWheelDeadzone) * settings->mouseWheelSensitivity / 100;
			} else if(mouse->dwx < -settings->mouseWheelDeadzone) {
                            hwheelChange = (mouse->dwx + settings->mouseWheelDeadzone) * settings->mouseWheelSensitivity / 100;
			}
                        if(hwheelChange != 0) {
                            mouse->hwheelPosition += hwheelChange;
                            mouse->device.send(EV_REL, REL_HWHEEL_HI_RES, hwheelChange);
                            int notches = mouse->hwheelPosition / mouse->WHEEL_HI_RES_PER_NOTCH;
                            if(notches != 0) {
				mouse->device.send(EV_REL, REL_HWHEEL, notches);
                                mouse->hwheelPosition = mouse->hwheelPosition % mouse->WHEEL_HI_RES_PER_NOTCH;
                            }
                        }

                        int wheelChange = 0;
			if(mouse->dwy > settings->mouseWheelDeadzone) {
                            wheelChange = -(mouse->dwy - settings->mouseWheelDeadzone) * settings->mouseWheelSensitivity / 100;
			} else if(mouse->dwy < -settings->mouseWheelDeadzone) {
                            wheelChange = -(mouse->dwy + settings->mouseWheelDeadzone) * settings->mouseWheelSensitivity / 100;
			}
                        if(wheelChange != 0) {
                            mouse->wheelPosition += wheelChange;
                            mouse->device.send(EV_REL, REL_WHEEL_HI_RES, wheelChange);
                            int notches = mouse->wheelPosition / mouse->WHEEL_HI_RES_PER_NOTCH;
                            if(notches != 0) {
				mouse->device.send(EV_REL, REL_WHEEL, notches);
                                mouse->wheelPosition = mouse->wheelPosition % mouse->WHEEL_HI_RES_PER_NOTCH;
                            }
                        }

			mouse->device.send(EV_SYN, 0, 0);
		} else {
			std::unique_lock<std::mutex> lk(mouse->mutex);
			mouse->signal.wait(lk);
		}
		std::this_thread::sleep_for(std::chrono::milliseconds(16));
	}
	delete mouse;
}

template<class T> std::vector<T> parseMultivalue(std::string const& str, std::function<T(std::string const&)> fn) {
  std::istringstream stream(str);
  std::vector<T> values;
  std::transform(std::istream_iterator<std::string>(stream), std::istream_iterator<std::string>(), std::back_inserter(values), fn);
  return values;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::passthrough(unsigned int code) {
	behaviors.at(code - FIRST_KEY).type = KeyBehavior::PASSTHROUGH;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::map(unsigned int code, unsigned int result) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::MAPPED;
	b.mapping = result;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::altmap(unsigned int code, bool* flag, unsigned int regular, unsigned int alternative) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::ALTMAPPED;
	b.mapping = regular;
	b.alternative = alternative;
	b.flag = flag;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::complex(unsigned int code, std::function<void(int)> function) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::COMPLEX;
	b.function = function;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::altcomplex(unsigned int code, bool* flag, unsigned int regular, std::function<void(int)> function) {
	auto& b = behaviors.at(code - FIRST_KEY);
        b.type = KeyBehavior::ALTCOMPLEX;
        b.mapping = regular;
        b.flag = flag;
        b.function = function;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::gpmap(unsigned int code, unsigned int gamepad) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPMAPPED;
	b.alternative = gamepad;
}
template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::gpmap2(unsigned int code, unsigned int gamepad, bool* self, LeftRight* pair) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPMAP2;
	b.alternative = gamepad;
	b.flag  = self;
	b.lr	= pair;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::gphat(unsigned int code, unsigned int gamepad) {
	auto& b = behaviors.at(code - FIRST_KEY);
	b.type = KeyBehavior::GPHAT;
	b.alternative = gamepad;
}

template<int FIRST_KEY, int LAST_KEY> void KeyBehaviors<FIRST_KEY, LAST_KEY>::handle(unsigned int code, int value) {
	if (code <FIRST_KEY || code >LAST_KEY) return;
	auto& kb = behaviors.at(code - FIRST_KEY);
	switch(kb.type) {
	case KeyBehavior::PASSTHROUGH:
		global.keyboard->send(EV_KEY, code, value);
		break;
	case KeyBehavior::MAPPED:
		global.keyboard->send(EV_KEY, kb.mapping, value);
		break;
	case KeyBehavior::ALTMAPPED:
		if(value==1)
			kb.pressed_as = *kb.flag;
		global.keyboard->send(EV_KEY, kb.pressed_as ? kb.alternative : kb.mapping, value);
		break; 
	case KeyBehavior::COMPLEX:
		kb.function(value);
		break;
	case KeyBehavior::ALTCOMPLEX:
		if(*kb.flag) {
                  kb.function(value);
                } else {
                  global.keyboard->send(EV_KEY, kb.mapping, value);
                }
		break;
	case KeyBehavior::GPMAPPED:
		if (global.settings.exportKeypad)
			global.keyboard->send(EV_KEY, code, value);
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_KEY, kb.alternative, value);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case KeyBehavior::GPMAP2:
		*kb.flag = (value==1);
		kb.lr->pressed = kb.lr->left || kb.lr->right;
		if (global.settings.exportKeypad)
			global.keyboard->send(EV_KEY, code, value);
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_KEY, kb.alternative, value);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
	case KeyBehavior::GPHAT:
		global.keyboard->send(EV_KEY, code, value);
		switch (kb.alternative) {
		case BTN_DPAD_UP:
			global.haty = -value;
			break;
		case BTN_DPAD_DOWN:
			global.haty = value;
			break;
		case BTN_DPAD_LEFT:
			global.hatx = -value;
			break;
		case BTN_DPAD_RIGHT:
			global.hatx = value;
			break;
		};
		if (global.settings.exportGamepad) {
			global.gamepad->send(EV_ABS, ABS_HAT0X, global.hatx*65535);
			global.gamepad->send(EV_ABS, ABS_HAT0Y, global.haty*65535);
			global.gamepad->send(EV_SYN, 0, 0);
		}
		break;
        }
}
